# -*- python-mode -*-
# -*- coding: UTF-8 -*-

## Copyright (C) 2012-2013  Daniel Pavel
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from __future__ import absolute_import, division, print_function, unicode_literals

from logging import getLogger, DEBUG as _DEBUG
_log = getLogger(__name__)
del getLogger

from .i18n import _
from . import hidpp10 as _hidpp10
from . import hidpp20 as _hidpp20
from .common import (
				bytes2int as _bytes2int,
				int2bytes as _int2bytes,
				NamedInts as _NamedInts,
				unpack as _unpack,
			)
from .settings import (
				KIND as _KIND,
				Setting as _Setting,
				RegisterRW as _RegisterRW,
				FeatureRW as _FeatureRW,
				BooleanValidator as _BooleanV,
				ChoicesValidator as _ChoicesV,
				RangeValidator as _RangeV,
			)

_DK = _hidpp10.DEVICE_KIND
_R = _hidpp10.REGISTERS
_F = _hidpp20.FEATURE

#
# pre-defined basic setting descriptors
#

def register_toggle(name, register,
					true_value=_BooleanV.default_true,
					false_value=_BooleanV.default_false,
					mask=_BooleanV.default_mask,
					label=None, description=None, device_kind=None):
	validator = _BooleanV(true_value=true_value, false_value=false_value, mask=mask)
	rw = _RegisterRW(register)
	return _Setting(name, rw, validator, label=label, description=description, device_kind=device_kind)


def register_choices(name, register, choices,
					kind=_KIND.choice,
					label=None, description=None, device_kind=None):
	assert choices
	validator = _ChoicesV(choices)
	rw = _RegisterRW(register)
	return _Setting(name, rw, validator, kind=kind, label=label, description=description, device_kind=device_kind)


def feature_toggle(name, feature,
					read_function_id=_FeatureRW.default_read_fnid,
					write_function_id=_FeatureRW.default_write_fnid,
					true_value=_BooleanV.default_true,
					false_value=_BooleanV.default_false,
					mask=_BooleanV.default_mask,
					label=None, description=None, device_kind=None):
	validator = _BooleanV(true_value=true_value, false_value=false_value, mask=mask)
	rw = _FeatureRW(feature, read_function_id, write_function_id)
	return _Setting(name, rw, validator, label=label, description=description, device_kind=device_kind)

def feature_choices(name, feature, choices,
					read_function_id, write_function_id,
					bytes_count=None,
					label=None, description=None, device_kind=None):
	assert choices
	validator = _ChoicesV(choices, bytes_count=bytes_count)
	rw = _FeatureRW(feature, read_function_id, write_function_id)
	return _Setting(name, rw, validator, kind=_KIND.choice, label=label, description=description, device_kind=device_kind)

def feature_choices_dynamic(name, feature, choices_callback,
					read_function_id, write_function_id,
					bytes_count=None,
					label=None, description=None, device_kind=None):
	# Proxy that obtains choices dynamically from a device
	def instantiate(device):
		# Obtain choices for this feature
		choices = choices_callback(device)
		setting = feature_choices(name, feature, choices,
						read_function_id, write_function_id,
						bytes_count=bytes_count,
						label=label, description=description, device_kind=device_kind)
		return setting(device)
	return instantiate

def feature_range(name, feature, min_value, max_value,
					read_function_id=_FeatureRW.default_read_fnid,
					write_function_id=_FeatureRW.default_write_fnid,
					rw=None,
					bytes_count=None,
					label=None, description=None, device_kind=None):
	validator = _RangeV(min_value, max_value, bytes_count=bytes_count)
	if rw is None:
		rw = _FeatureRW(feature, read_function_id, write_function_id)
	return _Setting(name, rw, validator, kind=_KIND.range, label=label, description=description, device_kind=device_kind)

#
# common strings for settings
#

_SMOOTH_SCROLL = ('smooth-scroll', _("Smooth Scrolling"),
							_("High-sensitivity mode for vertical scroll with the wheel."))
_LOW_RES_SCROLL = ('lowres-smooth-scroll', _("HID++ Scrolling"),
							_("HID++ mode for vertical scroll with the wheel."))

_HI_RES_SCROLL = ('hi-res-scroll', _("High Resolution Scrolling"),
							_("High-sensitivity mode for vertical scroll with the wheel."))
_HIRES_INV = ('hires-smooth-invert', _("High Resolution Wheel Invert"),
							_("High-sensitivity wheel invert mode for vertical scroll."))
_HIRES_RES = ('hires-smooth-resolution', _("Wheel Resolution"),
							_("High-sensitivity mode for vertical scroll with the wheel."))
_SIDE_SCROLL = ('side-scroll', _("Side Scrolling"),
							_("When disabled, pushing the wheel sideways sends custom button events\n"
							"instead of the standard side-scrolling events."))
_DPI = ('dpi', _("Sensitivity (DPI)"), None)
_POINTER_SPEED = ('pointer_speed', _("Sensitivity (Pointer Speed)"), None)
_FN_SWAP = ('fn-swap', _("Swap Fx function"),
							_("When set, the F1..F12 keys will activate their special function,\n"
						 	"and you must hold the FN key to activate their standard function.")
						 	+ '\n\n' +
						 	_("When unset, the F1..F12 keys will activate their standard function,\n"
						 	"and you must hold the FN key to activate their special function."))
_HAND_DETECTION = ('hand-detection', _("Hand Detection"),
							_("Turn on illumination when the hands hover over the keyboard."))
_BACKLIGHT = ('backlight', _("Backlight"),
				   _("Turn illumination on or off on keyboard."))

_SMART_SHIFT = ('smart-shift', _("Smart Shift"),
							_("Automatically switch the mouse wheel between ratchet and freespin mode.\n"
							"The mouse wheel is always free at 0, and always locked at 50"))
#
#
#

def _register_hand_detection(register=_R.keyboard_hand_detection,
					true_value=b'\x00\x00\x00', false_value=b'\x00\x00\x30', mask=b'\x00\x00\xFF'):
	return register_toggle(_HAND_DETECTION[0], register, true_value=true_value, false_value=false_value,
					label=_HAND_DETECTION[1], description=_HAND_DETECTION[2],
					device_kind=(_DK.keyboard,))

def _register_fn_swap(register=_R.keyboard_fn_swap, true_value=b'\x00\x01', mask=b'\x00\x01'):
	return register_toggle(_FN_SWAP[0], register, true_value=true_value, mask=mask,
					label=_FN_SWAP[1], description=_FN_SWAP[2],
					device_kind=(_DK.keyboard,))

def _register_smooth_scroll(register=_R.mouse_button_flags, true_value=0x40, mask=0x40):
	return register_toggle(_SMOOTH_SCROLL[0], register, true_value=true_value, mask=mask,
					label=_SMOOTH_SCROLL[1], description=_SMOOTH_SCROLL[2],
					device_kind=(_DK.mouse, _DK.trackball))

def _register_side_scroll(register=_R.mouse_button_flags, true_value=0x02, mask=0x02):
	return register_toggle(_SIDE_SCROLL[0], register, true_value=true_value, mask=mask,
 					label=_SIDE_SCROLL[1], description=_SIDE_SCROLL[2],
					device_kind=(_DK.mouse, _DK.trackball))

def _register_dpi(register=_R.mouse_dpi, choices=None):
	return register_choices(_DPI[0], register, choices,
					label=_DPI[1], description=_DPI[2],
					device_kind=(_DK.mouse, _DK.trackball))


def _feature_fn_swap():
	return feature_toggle(_FN_SWAP[0], _F.FN_INVERSION,
					label=_FN_SWAP[1], description=_FN_SWAP[2],
					device_kind=(_DK.keyboard,))

def _feature_new_fn_swap():
	return feature_toggle(_FN_SWAP[0], _F.NEW_FN_INVERSION,
					label=_FN_SWAP[1], description=_FN_SWAP[2],
					device_kind=(_DK.keyboard,))

def _feature_k375s_fn_swap():
	return feature_toggle(_FN_SWAP[0], _F.K375S_FN_INVERSION,
					label=_FN_SWAP[1], description=_FN_SWAP[2],
					device_kind=(_DK.keyboard,))

# FIXME: This will enable all supported backlight settings, we should allow the users to select which settings they want to enable.
def _feature_backlight2():
	return feature_toggle(_BACKLIGHT[0], _F.BACKLIGHT2,
						  label=_BACKLIGHT[1], description=_BACKLIGHT[2],
						  device_kind=(_DK.keyboard,))

def _feature_hi_res_scroll():
	return feature_toggle(_HI_RES_SCROLL[0], _F.HI_RES_SCROLLING,
					label=_HI_RES_SCROLL[1], description=_HI_RES_SCROLL[2],
					device_kind=(_DK.mouse, _DK.trackball))

def _feature_lowres_smooth_scroll():
	return feature_toggle(_LOW_RES_SCROLL[0], _F.LOWRES_WHEEL,
					label=_LOW_RES_SCROLL[1], description=_LOW_RES_SCROLL[2],
					device_kind=(_DK.mouse, _DK.trackball))
def _feature_hires_smooth_invert():
	return feature_toggle(_HIRES_INV[0], _F.HIRES_WHEEL,
					read_function_id=0x10,
					write_function_id=0x20,
					true_value=0x04, mask=0x04,
					label=_HIRES_INV[1], description=_HIRES_INV[2],
					device_kind=(_DK.mouse, _DK.trackball))

def _feature_hires_smooth_resolution():
	return feature_toggle(_HIRES_RES[0], _F.HIRES_WHEEL,
					read_function_id=0x10,
					write_function_id=0x20,
					true_value=0x02, mask=0x02,
					label=_HIRES_RES[1], description=_HIRES_RES[2],
					device_kind=(_DK.mouse, _DK.trackball))

def _feature_smart_shift():
	_MIN_SMART_SHIFT_VALUE = 0
	_MAX_SMART_SHIFT_VALUE = 50
	class _SmartShiftRW(_FeatureRW):
		def __init__(self, feature):
			super(_SmartShiftRW, self).__init__(feature)

		def read(self, device):
			value = super(_SmartShiftRW, self).read(device)
			if _bytes2int(value[0:1]) == 1:
				# Mode = Freespin, map to minimum
				return _int2bytes(_MIN_SMART_SHIFT_VALUE, count=1)
			else:
				# Mode = smart shift, map to the value, capped at maximum
				threshold = min(_bytes2int(value[1:2]), _MAX_SMART_SHIFT_VALUE)
				return _int2bytes(threshold, count=1)

		def write(self, device, data_bytes):
			threshold = _bytes2int(data_bytes)
			# Freespin at minimum
			mode = 1 if threshold == _MIN_SMART_SHIFT_VALUE else 2

			# Ratchet at maximum
			if threshold == _MAX_SMART_SHIFT_VALUE:
				threshold = 255

			data = _int2bytes(mode, count=1) + _int2bytes(threshold, count=1) * 2
			return super(_SmartShiftRW, self).write(device, data)

	return feature_range(_SMART_SHIFT[0], _F.SMART_SHIFT,
	                _MIN_SMART_SHIFT_VALUE, _MAX_SMART_SHIFT_VALUE,
					bytes_count=1,
					rw=_SmartShiftRW(_F.SMART_SHIFT),
					label=_SMART_SHIFT[1], description=_SMART_SHIFT[2],
					device_kind=(_DK.mouse, _DK.trackball))

def _feature_adjustable_dpi_choices(device):
	# [1] getSensorDpiList(sensorIdx)
	reply = device.feature_request(_F.ADJUSTABLE_DPI, 0x10)
	# Should not happen, but might happen when the user unplugs device while the
	# query is being executed. TODO retry logic?
	assert reply, 'Oops, DPI list cannot be retrieved!'
	dpi_list = []
	step = None
	for val in _unpack('!7H', reply[1:1+14]):
		if val == 0:
			break
		if val >> 13 == 0b111:
			assert step is None and len(dpi_list) == 1, \
					'Invalid DPI list item: %r' % val
			step = val & 0x1fff
		else:
			dpi_list.append(val)
	if step:
		assert len(dpi_list) == 2, 'Invalid DPI list range: %r' % dpi_list
		dpi_list = range(dpi_list[0], dpi_list[1] + 1, step)
	return _NamedInts.list(dpi_list)

def _feature_adjustable_dpi():
	"""Pointer Speed feature"""
	# Assume sensorIdx 0 (there is only one sensor)
	# [2] getSensorDpi(sensorIdx) -> sensorIdx, dpiMSB, dpiLSB
	# [3] setSensorDpi(sensorIdx, dpi)
	return feature_choices_dynamic(_DPI[0], _F.ADJUSTABLE_DPI,
					_feature_adjustable_dpi_choices,
					read_function_id=0x20,
					write_function_id=0x30,
					bytes_count=3,
					label=_DPI[1], description=_DPI[2],
					device_kind=(_DK.mouse, _DK.trackball))

def _feature_pointer_speed():
	"""Pointer Speed feature"""
	# min and max values taken from usb traces of Win software
	return feature_range(_POINTER_SPEED[0], _F.POINTER_SPEED, 0x002e, 0x01ff,
					read_function_id=0x0,
					write_function_id=0x10,
					bytes_count=2,
					label=_POINTER_SPEED[1], description=_POINTER_SPEED[2],
					device_kind=(_DK.mouse, _DK.trackball))
#
#
#

from collections import namedtuple
_SETTINGS_LIST = namedtuple('_SETTINGS_LIST', [
					'fn_swap',
					'new_fn_swap',
					'k375s_fn_swap',
					'smooth_scroll',
					'hi_res_scroll',
					'lowres_smooth_scroll',
					'hires_smooth_invert',
					'hires_smooth_resolution',
					'side_scroll',
					'dpi',
					'pointer_speed',
					'hand_detection',
					'backlight',
					'typing_illumination',
					'smart_shift',
					])
del namedtuple

RegisterSettings = _SETTINGS_LIST(
				fn_swap=_register_fn_swap,
				new_fn_swap=None,
				k375s_fn_swap=None,
				smooth_scroll=_register_smooth_scroll,
				hi_res_scroll=None,
				lowres_smooth_scroll=None,
				hires_smooth_invert=None,
				hires_smooth_resolution=None,
				side_scroll=_register_side_scroll,
				dpi=_register_dpi,
				pointer_speed=None,
				hand_detection=_register_hand_detection,
				backlight=None,
				typing_illumination=None,
				smart_shift=None,
			)
FeatureSettings =  _SETTINGS_LIST(
				fn_swap=_feature_fn_swap,
				new_fn_swap=_feature_new_fn_swap,
				k375s_fn_swap=_feature_k375s_fn_swap,
				smooth_scroll=None,
				hi_res_scroll=_feature_hi_res_scroll,
				lowres_smooth_scroll=_feature_lowres_smooth_scroll,
				hires_smooth_invert=_feature_hires_smooth_invert,
				hires_smooth_resolution=_feature_hires_smooth_resolution,
				side_scroll=None,
				dpi=_feature_adjustable_dpi,
				pointer_speed=_feature_pointer_speed,
				hand_detection=None,
				backlight=_feature_backlight2,
				typing_illumination=None,
				smart_shift=_feature_smart_shift,
			)

del _SETTINGS_LIST

#
#
#

def check_feature_settings(device, already_known):
	"""Try to auto-detect device settings by the HID++ 2.0 features they have."""
	if device.features is None or not device.online:
		return
	if device.protocol and device.protocol < 2.0:
		return

	def check_feature(name, featureId, field_name=None):
		"""
		:param name: user-visible setting name.
		:param featureId: the numeric Feature ID for this setting.
		:param field_name: override the FeatureSettings name if it is
		different from the user-visible setting name. Useful if there
		are multiple features for the same setting.
		"""
		if not featureId in device.features:
			return
		if any(s.name == name for s in already_known):
			return
		if not field_name:
			# Convert user-visible settings name for FeatureSettings
			field_name = name.replace('-', '_')
		feature = getattr(FeatureSettings, field_name)()

		try:
			detected = feature(device)
			if _log.isEnabledFor(_DEBUG):
			    _log.debug("check_feature[%s] detected %s", featureId, detected)
			already_known.append(detected)
		except Exception as reason:
			_log.error("check_feature[%s] inconsistent feature %s", featureId, reason)

	check_feature(_HI_RES_SCROLL[0], _F.HI_RES_SCROLLING)
	check_feature(_LOW_RES_SCROLL[0], _F.LOWRES_WHEEL)
	check_feature(_HIRES_INV[0],     _F.HIRES_WHEEL, "hires_smooth_invert")
	check_feature(_HIRES_RES[0],     _F.HIRES_WHEEL, "hires_smooth_resolution")
	check_feature(_FN_SWAP[0],       _F.FN_INVERSION)
	check_feature(_FN_SWAP[0],       _F.NEW_FN_INVERSION, 'new_fn_swap')
	check_feature(_FN_SWAP[0],       _F.K375S_FN_INVERSION, 'k375s_fn_swap')
	check_feature(_DPI[0],           _F.ADJUSTABLE_DPI)
	check_feature(_POINTER_SPEED[0], _F.POINTER_SPEED)
	check_feature(_SMART_SHIFT[0],   _F.SMART_SHIFT)
	check_feature(_BACKLIGHT[0],   	 _F.BACKLIGHT2)



#!/usr/bin/env python3
# -*- python-mode -*-
# -*- coding: UTF-8 -*-

## Copyright (C) 2012-2013  Daniel Pavel
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from __future__ import absolute_import, division, print_function, unicode_literals

from logging import getLogger, DEBUG as _DEBUG
_log = getLogger(__name__)
del getLogger

from threading import Thread as _Thread

try:
	from Queue import Queue as _Queue
except ImportError:
	from queue import Queue as _Queue

#
#
#

class TaskRunner(_Thread):
	def __init__(self, name):
		super(TaskRunner, self).__init__(name=name)
		self.daemon = True
		self.queue = _Queue(16)
		self.alive = False

	def __call__(self, function, *args, **kwargs):
		task = (function, args, kwargs)
		self.queue.put(task)

	def stop(self):
		self.alive = False
		self.queue.put(None)

	def run(self):
		self.alive = True

		if _log.isEnabledFor(_DEBUG):
			_log.debug("started")

		while self.alive:
			task = self.queue.get()
			if task:
				function, args, kwargs = task
				assert function
				try:
					function(*args, **kwargs)
				except:
					_log.exception("calling %s", function)

		if _log.isEnabledFor(_DEBUG):
			_log.debug("stopped")



# -*- python-mode -*-
# -*- coding: UTF-8 -*-

## Copyright (C) 2012-2013  Daniel Pavel
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from __future__ import absolute_import, division, print_function, unicode_literals

from time import time as _timestamp

from logging import getLogger, DEBUG as _DEBUG
_log = getLogger(__name__)
del getLogger


from .i18n import _, ngettext
from .common import NamedInts as _NamedInts, NamedInt as _NamedInt
from . import hidpp10 as _hidpp10
from . import hidpp20 as _hidpp20

_R = _hidpp10.REGISTERS

#
#
#

ALERT = _NamedInts(NONE=0x00, NOTIFICATION=0x01, SHOW_WINDOW=0x02, ATTENTION=0x04, ALL=0xFF)

KEYS = _NamedInts(
				BATTERY_LEVEL=1,
				BATTERY_CHARGING=2,
				BATTERY_STATUS=3,
				LIGHT_LEVEL=4,
				LINK_ENCRYPTED=5,
				NOTIFICATION_FLAGS=6,
				ERROR=7,
			)

# If the battery charge is under this percentage, trigger an attention event
# (blink systray icon/notification/whatever).
_BATTERY_ATTENTION_LEVEL = 5

# If no updates have been receiver from the device for a while, ping the device
# and update it status accordingly.
# _STATUS_TIMEOUT = 5 * 60  # seconds
_LONG_SLEEP = 15 * 60  # seconds

#
#
#

def attach_to(device, changed_callback):
	assert device
	assert changed_callback

	if not hasattr(device, 'status') or device.status is None:
		if device.kind is None:
			device.status = ReceiverStatus(device, changed_callback)
		else:
			device.status = DeviceStatus(device, changed_callback)

#
#
#

class ReceiverStatus(dict):
	"""The 'runtime' status of a receiver, mostly about the pairing process --
	is the pairing lock open or closed, any pairing errors, etc.
	"""
	def __init__(self, receiver, changed_callback):
		assert receiver
		self._receiver = receiver

		assert changed_callback
		self._changed_callback = changed_callback

		# self.updated = 0

		self.lock_open = False
		self.new_device = None

		self[KEYS.ERROR] = None

	def __str__(self):
		count = len(self._receiver)
		return (_("No paired devices.") if count == 0 else
			ngettext("%(count)s paired device.", "%(count)s paired devices.", count) % { 'count': count })
	__unicode__ = __str__

	def changed(self, alert=ALERT.NOTIFICATION, reason=None):
		# self.updated = _timestamp()
		self._changed_callback(self._receiver, alert=alert, reason=reason)

	# def poll(self, timestamp):
	# 	r = self._receiver
	# 	assert r
	#
	# 	if _log.isEnabledFor(_DEBUG):
	# 		_log.debug("polling status of %s", r)
	#
	# 	# make sure to read some stuff that may be read later by the UI
	# 	r.serial, r.firmware, None
	#
	# 	# get an update of the notification flags
	# 	# self[KEYS.NOTIFICATION_FLAGS] = _hidpp10.get_notification_flags(r)

#
#
#

class DeviceStatus(dict):
	"""Holds the 'runtime' status of a peripheral -- things like
	active/inactive, battery charge, lux, etc. It updates them mostly by
	processing incoming notification events from the device itself.
	"""
	def __init__(self, device, changed_callback):
		assert device
		self._device = device

		assert changed_callback
		self._changed_callback = changed_callback

		# is the device active?
		self._active = None

		# timestamp of when this status object was last updated
		self.updated = 0

	def to_string(self):
		def _items():
			comma = False

			battery_level = self.get(KEYS.BATTERY_LEVEL)
			if battery_level is not None:
				if isinstance(battery_level, _NamedInt):
					yield _("Battery: %(level)s") % { 'level': _(str(battery_level)) }
				else:
					yield _("Battery: %(percent)d%%") % { 'percent': battery_level }

				battery_status = self.get(KEYS.BATTERY_STATUS)
				if battery_status is not None:
					yield ' (%s)' % _(str(battery_status))

				comma = True

			light_level = self.get(KEYS.LIGHT_LEVEL)
			if light_level is not None:
				if comma: yield ', '
				yield _("Lighting: %(level)s lux") % { 'level': light_level }

		return ''.join(i for i in _items())

	def __repr__(self):
		return '{' +  ', '.join('\'%s\': %r' % (k, v) for k, v in self.items()) + '}'

	def __bool__(self):
		return bool(self._active)
	__nonzero__ = __bool__

	def set_battery_info(self, level, status, timestamp=None):
		if _log.isEnabledFor(_DEBUG):
			_log.debug("%s: battery %s, %s", self._device, level, status)

		if level is None:
			# Some notifications may come with no battery level info, just
			# charging state info, so do our best to infer a level (even if it is just the last level)
			# It is not always possible to do this well
			if status == _hidpp20.BATTERY_STATUS.full:
				level = _hidpp10.BATTERY_APPOX.full
			elif status in (_hidpp20.BATTERY_STATUS.almost_full, _hidpp20.BATTERY_STATUS.recharging):
				level = _hidpp10.BATTERY_APPOX.good
			elif status == _hidpp20.BATTERY_STATUS.slow_recharge:
				level = _hidpp10.BATTERY_APPOX.low
			else:
				level = self.get(KEYS.BATTERY_LEVEL)
		else:
			assert isinstance(level, int)

		# TODO: this is also executed when pressing Fn+F7 on K800.
		old_level, self[KEYS.BATTERY_LEVEL] = self.get(KEYS.BATTERY_LEVEL), level
		old_status, self[KEYS.BATTERY_STATUS] = self.get(KEYS.BATTERY_STATUS), status

		charging = status in (_hidpp20.BATTERY_STATUS.recharging, _hidpp20.BATTERY_STATUS.almost_full,
				      _hidpp20.BATTERY_STATUS.full, _hidpp20.BATTERY_STATUS.slow_recharge)
		old_charging, self[KEYS.BATTERY_CHARGING] = self.get(KEYS.BATTERY_CHARGING), charging

		changed = old_level != level or old_status != status or old_charging != charging
		alert, reason = ALERT.NONE, None

		if _hidpp20.BATTERY_OK(status) and ( level is None or level > _BATTERY_ATTENTION_LEVEL ):
			self[KEYS.ERROR] = None
		else:
			_log.warn("%s: battery %d%%, ALERT %s", self._device, level, status)
			if self.get(KEYS.ERROR) != status:
				self[KEYS.ERROR] = status
				# only show the notification once
				alert = ALERT.NOTIFICATION | ALERT.ATTENTION
			if isinstance(level, _NamedInt):
				reason = _("Battery: %(level)s (%(status)s)") % { 'level': _(level), 'status': _(status) }
			else:
				reason = _("Battery: %(percent)d%% (%(status)s)") % { 'percent': level, 'status': status.name }

		if changed or reason:
			# update the leds on the device, if any
			_hidpp10.set_3leds(self._device, level, charging=charging, warning=bool(alert))
			self.changed(active=True, alert=alert, reason=reason, timestamp=timestamp)

	def read_battery(self, timestamp=None):
		if self._active:
			d = self._device
			assert d

			if d.protocol < 2.0:
				battery = _hidpp10.get_battery(d)
			else:
				battery = _hidpp20.get_battery(d)

			# Really unnecessary, if the device has SOLAR_DASHBOARD it should be
			# broadcasting it's battery status anyway, it will just take a little while.
			# However, when the device has just been detected, it will not show
			# any battery status for a while (broadcasts happen every 90 seconds).
			if battery is None and d.features and _hidpp20.FEATURE.SOLAR_DASHBOARD in d.features:
				d.feature_request(_hidpp20.FEATURE.SOLAR_DASHBOARD, 0x00, 1, 1)
				return

			if battery is not None:
				level, status = battery
				self.set_battery_info(level, status)
			elif KEYS.BATTERY_STATUS in self:
				self[KEYS.BATTERY_STATUS] = None
				self[KEYS.BATTERY_CHARGING] = None
				self.changed()

	def changed(self, active=None, alert=ALERT.NONE, reason=None, timestamp=None):
		assert self._changed_callback
		d = self._device
		# assert d  # may be invalid when processing the 'unpaired' notification
		timestamp = timestamp or _timestamp()

		if active is not None:
			d.online = active
			was_active, self._active = self._active, active
			if active:
				if not was_active:
					# Make sure to set notification flags on the device, they
					# get cleared when the device is turned off (but not when the device
					# goes idle, and we can't tell the difference right now).
					if d.protocol < 2.0:
						self[KEYS.NOTIFICATION_FLAGS] = d.enable_notifications()

					# If we've been inactive for a long time, forget anything
					# about the battery.
					if self.updated > 0 and timestamp - self.updated > _LONG_SLEEP:
						self[KEYS.BATTERY_LEVEL] = None
						self[KEYS.BATTERY_STATUS] = None
						self[KEYS.BATTERY_CHARGING] = None

					# Devices lose configuration when they are turned off,
					# make sure they're up-to-date.
					if _log.isEnabledFor(_DEBUG):
						_log.debug("%s pushing device settings %s", d, d.settings)
					for s in d.settings:
						s.apply()

					if self.get(KEYS.BATTERY_LEVEL) is None:
						self.read_battery(timestamp)
			else:
				if was_active:
					battery = self.get(KEYS.BATTERY_LEVEL)
					self.clear()
					# If we had a known battery level before, assume it's not going
					# to change much while the device is offline.
					if battery is not None:
						self[KEYS.BATTERY_LEVEL] = battery

		if self.updated == 0 and active == True:
			# if the device is active on the very first status notification,
			# (meaning just when the program started or a new receiver was just
			# detected), pop-up a notification about it
			alert |= ALERT.NOTIFICATION
		self.updated = timestamp

		# if _log.isEnabledFor(_DEBUG):
		# 	_log.debug("device %d changed: active=%s %s", d.number, self._active, dict(self))
		self._changed_callback(d, alert, reason)

	# def poll(self, timestamp):
	# 	d = self._device
	# 	if not d:
	# 		_log.error("polling status of invalid device")
	# 		return
	#
	# 	if self._active:
	# 		if _log.isEnabledFor(_DEBUG):
	# 			_log.debug("polling status of %s", d)
	#
	# 		# read these from the device, the UI may need them later
	# 		d.protocol, d.serial, d.firmware, d.kind, d.name, d.settings, None
	#
	# 		# make sure we know all the features of the device
	# 		# if d.features:
	# 		# 	d.features[:]
	#
	# 		# devices may go out-of-range while still active, or the computer
	# 		# may go to sleep and wake up without the devices available
	# 		if timestamp - self.updated > _STATUS_TIMEOUT:
	# 			if d.ping():
	# 				timestamp = self.updated = _timestamp()
	# 			else:
	# 				self.changed(active=False, reason='out of range')
	#
	# 		# if still active, make sure we know the battery level
	# 		if KEYS.BATTERY_LEVEL not in self:
	# 			self.read_battery(timestamp)
	#
	# 	elif timestamp - self.updated > _STATUS_TIMEOUT:
	# 		if d.ping():
	# 			self.changed(active=True)
	# 		else:
	# 			self.updated = _timestamp()



# -*- python-mode -*-
# -*- coding: UTF-8 -*-

## Copyright (C) 2012-2013  Daniel Pavel
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from __future__ import absolute_import, division, print_function, unicode_literals


from logitech_receiver import (
				hidpp10 as _hidpp10,
				hidpp20 as _hidpp20,
				special_keys as _special_keys,
			)


def _print_receiver(receiver):
	paired_count = receiver.count()

	print (receiver.name)
	print ('  Device path  :', receiver.path)
	print ('  USB id       : 046d:%s' % receiver.product_id)
	print ('  Serial       :', receiver.serial)
	if receiver.firmware:
		for f in receiver.firmware:
			print ('    %-11s: %s' % (f.kind, f.version))

	print ('  Has', paired_count, 'paired device(s) out of a maximum of %d.' % receiver.max_devices)
	if receiver.remaining_pairings() and receiver.remaining_pairings() >= 0 :
		print ('  Has %d successful pairing(s) remaining.' % receiver.remaining_pairings() )

	notification_flags = _hidpp10.get_notification_flags(receiver)
	if notification_flags is not None:
		if notification_flags:
			notification_names = _hidpp10.NOTIFICATION_FLAG.flag_names(notification_flags)
			print ('  Notifications: %s (0x%06X)' % (', '.join(notification_names), notification_flags))
		else:
			print ('  Notifications: (none)')

	activity = receiver.read_register(_hidpp10.REGISTERS.devices_activity)
	if activity:
		activity = [(d, ord(activity[d - 1:d])) for d in range(1, receiver.max_devices)]
		activity_text = ', '.join(('%d=%d' % (d, a)) for d, a in activity if a > 0)
		print ('  Device activity counters:', activity_text or '(empty)')


def _print_device(dev):
	assert dev
	# check if the device is online
	dev.ping()

	print ('  %d: %s' % (dev.number, dev.name))
	print ('     Codename     :', dev.codename)
	print ('     Kind         :', dev.kind)
	print ('     Wireless PID :', dev.wpid)
	if dev.protocol:
		print ('     Protocol     : HID++ %1.1f' % dev.protocol)
	else:
		print ('     Protocol     : unknown (device is offline)')
	if dev.polling_rate:
		print ('     Polling rate :', dev.polling_rate, 'ms (%dHz)' % (1000 // dev.polling_rate))
	print ('     Serial number:', dev.serial)
	if dev.firmware:
		for fw in dev.firmware:
			print ('       %11s:' % fw.kind, (fw.name + ' ' + fw.version).strip())

	if dev.power_switch_location:
		print ('     The power switch is located on the %s.' % dev.power_switch_location)

	if dev.online:
		notification_flags = _hidpp10.get_notification_flags(dev)
		if notification_flags is not None:
			if notification_flags:
				notification_names = _hidpp10.NOTIFICATION_FLAG.flag_names(notification_flags)
				print ('     Notifications: %s (0x%06X).' % (', '.join(notification_names), notification_flags))
			else:
				print ('     Notifications: (none).')

	if dev.online and dev.features:
		print ('     Supports %d HID++ 2.0 features:' % len(dev.features))
		for index, feature in enumerate(dev.features):
			feature = dev.features[index]
			flags = dev.request(0x0000, feature.bytes(2))
			flags = 0 if flags is None else ord(flags[1:2])
			flags = _hidpp20.FEATURE_FLAG.flag_names(flags)
			print ('        %2d: %-22s {%04X}   %s' % (index, feature, feature, ', '.join(flags)))
			if feature == _hidpp20.FEATURE.HIRES_WHEEL:
				wheel = _hidpp20.get_hires_wheel(dev)
				if wheel:
					multi, has_invert, has_switch, inv, res, target, ratchet = wheel
					print("            Multiplier: %s" % multi)
					if has_invert:
						print("            Has invert")
						if inv:
							print("              Inverse wheel motion")
						else:
							print("              Normal wheel motion")
					if has_switch:
						print("            Has ratchet switch")
						if ratchet:
							print("              Normal wheel mode")
						else:
							print("              Free wheel mode")
					if res:
						print("            High resolution mode")
					else:
						print("            Low resolution mode")
					if target:
						print("            HID++ notification")
					else:
						print("            HID notification")
			if feature == _hidpp20.FEATURE.MOUSE_POINTER:
				mouse_pointer = _hidpp20.get_mouse_pointer_info(dev)
				if mouse_pointer:
					print("            DPI: %s" % mouse_pointer['dpi'])
					print("            Acceleration: %s" % mouse_pointer['acceleration'])
					if mouse_pointer['suggest_os_ballistics']:
						print("            Use OS ballistics")
					else:
						print("            Override OS ballistics")
					if mouse_pointer['suggest_vertical_orientation']:
						print("            Provide vertical tuning, trackball")
					else:
						print("            No vertical tuning, standard mice")
			if feature == _hidpp20.FEATURE.VERTICAL_SCROLLING:
				vertical_scrolling_info = _hidpp20.get_vertical_scrolling_info(dev)
				if vertical_scrolling_info:
					print("            Roller type: %s" % vertical_scrolling_info['roller'])
					print("            Ratchet per turn: %s" % vertical_scrolling_info['ratchet'])
					print("            Scroll lines: %s" % vertical_scrolling_info['lines'])
			if feature == _hidpp20.FEATURE.HI_RES_SCROLLING:
				scrolling_mode, scrolling_resolution = _hidpp20.get_hi_res_scrolling_info(dev)
				if scrolling_mode:
					print("            Hi-res scrolling enabled")
				else:
					print("            Hi-res scrolling disabled")
				if scrolling_resolution:
					print("            Hi-res scrolling multiplier: %s" % scrolling_resolution)
			if feature == _hidpp20.FEATURE.POINTER_SPEED:
				pointer_speed = _hidpp20.get_pointer_speed_info(dev)
				if pointer_speed:
					print("            Pointer Speed: %s" % pointer_speed)
			if feature == _hidpp20.FEATURE.LOWRES_WHEEL:
				wheel_status = _hidpp20.get_lowres_wheel_status(dev)
				if wheel_status:
					print("            Wheel Reports: %s" % wheel_status)

	if dev.online and dev.keys:
		print ('     Has %d reprogrammable keys:' % len(dev.keys))
		for k in dev.keys:
			flags = _special_keys.KEY_FLAG.flag_names(k.flags)
			# TODO: add here additional variants for other REPROG_CONTROLS
			if dev.keys.keyversion == 1:
				print ('        %2d: %-26s => %-27s   %s' % (k.index, k.key, k.task, ', '.join(flags)))
			if dev.keys.keyversion == 4:
				print ('        %2d: %-26s, default: %-27s => %-26s' % (k.index, k.key, k.task, k.remapped))
				print ('             %s, pos:%d, group:%1d, gmask:%d' % ( ', '.join(flags), k.pos, k.group, k.group_mask))
	if dev.online:
		battery = _hidpp20.get_battery(dev)
		if battery is None:
			battery = _hidpp10.get_battery(dev)
		if battery is not None:
			from logitech_receiver.common import NamedInt as _NamedInt
			level, status = battery
			if level is not None:
				if isinstance(level, _NamedInt):
					text = str(level)
				else:
					text = '%d%%' % level
			else:
				text = 'N/A'
			print ('     Battery: %s, %s.' % (text, status))
		else:
			battery_voltage = _hidpp20.get_voltage(dev)
			if battery_voltage :
				(voltage, charging, charge_sts, charge_lvl, charge_type) = battery_voltage
				print ('     Battery: %smV, %s.' % (voltage, 'Charging' if charging else 'Discharging'))
			else:
				print ('     Battery status unavailable.')
	else:
		print ('     Battery: unknown (device is offline).')


def run(receivers, args, find_receiver, find_device):
	assert receivers
	assert args.device

	device_name = args.device.lower()

	if device_name == 'all':
		for r in receivers:
			_print_receiver(r)
			count = r.count()
			if count:
				for dev in r:
					print ('')
					_print_device(dev)
					count -= 1
					if not count:
						break
			print ('')
		return

	dev = find_receiver(receivers, device_name)
	if dev:
		_print_receiver(dev)
		return

	dev = find_device(receivers, device_name)
	assert dev
	_print_device(dev)



#!/usr/bin/env python3
# -*- python-mode -*-
# -*- coding: UTF-8 -*-

## Copyright (C) 2012-2013  Daniel Pavel
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from __future__ import absolute_import, division, print_function, unicode_literals

import importlib


from solaar import __version__, NAME
import solaar.i18n as _i18n
import solaar.cli as _cli

#
#
#

def _require(module, os_package, gi=None, gi_package=None, gi_version=None):
	try:
		if gi is not None:
			gi.require_version(gi_package,gi_version)
		return importlib.import_module(module)
	except (ImportError, ValueError):
		import sys
		sys.exit("%s: missing required system package %s" % (NAME, os_package))

prefer_symbolic_battery_icons = False

def _parse_arguments():
	import argparse
	arg_parser = argparse.ArgumentParser(prog=NAME.lower())
	arg_parser.add_argument('-d', '--debug', action='count', default=0,
							help='print logging messages, for debugging purposes (may be repeated for extra verbosity)')
	arg_parser.add_argument('-D', '--hidraw', action='store', dest='hidraw_path', metavar='PATH',
							help='unifying receiver to use; the first detected receiver if unspecified. Example: /dev/hidraw2')
	arg_parser.add_argument('--restart-on-wake-up', action='store_true',
							help='restart Solaar on sleep wake-up (experimental)')
	arg_parser.add_argument('-w', '--window', choices=('show','hide','only'), help='start with window showing / hidden / only (no tray icon)')
	arg_parser.add_argument('-b', '--battery-icons', choices=('regular','symbolic'), help='prefer regular / symbolic icons')
	arg_parser.add_argument('-V', '--version', action='version', version='%(prog)s ' + __version__)
	arg_parser.add_argument('--help-actions', action='store_true',
							help='print help for the optional actions')
	arg_parser.add_argument('action', nargs=argparse.REMAINDER, choices=_cli.actions,
							help='optional actions to perform')

	args = arg_parser.parse_args()

	if args.help_actions:
		_cli.print_help()
		return

	if args.window is None:
		args.window = 'show' # default behaviour is to show main window

	global prefer_symbolic_battery_icons
	prefer_symbolic_battery_icons = True if args.battery_icons == 'symbolic' else False

	import logging
	if args.debug > 0:
		log_level = logging.WARNING - 10 * args.debug
		log_format='%(asctime)s,%(msecs)03d %(levelname)8s [%(threadName)s] %(name)s: %(message)s'
		logging.basicConfig(level=max(log_level, logging.DEBUG), format=log_format, datefmt='%H:%M:%S')
	else:
		logging.root.addHandler(logging.NullHandler())
		logging.root.setLevel(logging.ERROR)

	if not args.action:
		if logging.root.isEnabledFor(logging.INFO):
			logging.info("language %s (%s), translations path %s", _i18n.language, _i18n.encoding, _i18n.path)

	return args


def main():
	_require('pyudev', 'python3-pyudev')

	# handle ^C in console
	import signal
	signal.signal(signal.SIGINT, signal.SIG_DFL)

	args = _parse_arguments()
	if not args: return
	if args.action:
		# if any argument, run comandline and exit
		return _cli.run(args.action, args.hidraw_path)

	gi = _require('gi', 'python3-gi or python3-gobject')
	_require('gi.repository.Gtk', 'gir1.2-gtk-3.0', gi, 'Gtk', '3.0')

	try:
		import solaar.ui as ui
		import solaar.listener as listener
		listener.setup_scanner(ui.status_changed, ui.error_dialog)

		import solaar.upower as _upower
		if args.restart_on_wake_up:
			_upower.watch(listener.start_all, listener.stop_all)
		else:
			_upower.watch(lambda: listener.ping_all(True))

		# main UI event loop
		ui.run_loop(listener.start_all, listener.stop_all, args.window!='only', args.window!='hide')
	except Exception as e:
		import sys
		sys.exit('%s: error: %s' % (NAME.lower(), e))


if __name__ == '__main__':
	main()



# -*- python-mode -*-
# -*- coding: UTF-8 -*-

## Copyright (C) 2012-2013  Daniel Pavel
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

import os as _os
import os.path as _path
from json import load as _json_load, dump as _json_save

from logging import getLogger, DEBUG as _DEBUG, INFO as _INFO
_log = getLogger(__name__)
del getLogger

_XDG_CONFIG_HOME = _os.environ.get('XDG_CONFIG_HOME') or _path.expanduser(_path.join('~', '.config'))
_file_path = _path.join(_XDG_CONFIG_HOME, 'solaar', 'config.json')


from solaar import __version__
_KEY_VERSION = '_version'
_KEY_NAME = '_name'
_configuration = {}



def _load():
	if _path.isfile(_file_path):
		loaded_configuration = {}
		try:
			with open(_file_path, 'r') as config_file:
				loaded_configuration = _json_load(config_file)
		except:
			_log.error("failed to load from %s", _file_path)

		# loaded_configuration.update(_configuration)
		_configuration.clear()
		_configuration.update(loaded_configuration)

	if _log.isEnabledFor(_DEBUG):
		_log.debug("load => %s", _configuration)

	_cleanup(_configuration)
	_configuration[_KEY_VERSION] = __version__
	return _configuration


def save():
	# don't save if the configuration hasn't been loaded
	if _KEY_VERSION not in _configuration:
		return

	dirname = _os.path.dirname(_file_path)
	if not _path.isdir(dirname):
		try:
			_os.makedirs(dirname)
		except:
			_log.error("failed to create %s", dirname)
			return False

	_cleanup(_configuration)

	try:
		with open(_file_path, 'w') as config_file:
			_json_save(_configuration, config_file, skipkeys=True, indent=2, sort_keys=True)

		if _log.isEnabledFor(_INFO):
			_log.info("saved %s to %s", _configuration, _file_path)
		return True
	except:
		_log.error("failed to save to %s", _file_path)


def _cleanup(d):
	# remove None values from the dict
	for key in list(d.keys()):
		value = d.get(key)
		if value is None:
			del d[key]
		elif isinstance(value, dict):
			_cleanup(value)


def _device_key(device):
	return '%s:%s' % (device.wpid, device.serial)


class _DeviceEntry(dict):
	def __init__(self, *args, **kwargs):
		super(_DeviceEntry, self).__init__(*args, **kwargs)

	def __setitem__(self, key, value):
		super(_DeviceEntry, self).__setitem__(key, value)
		save()


def _device_entry(device):
	if not _configuration:
		_load()

	device_key = _device_key(device)
	c = _configuration.get(device_key) or {}

	if not isinstance(c, _DeviceEntry):
		c[_KEY_NAME] = device.name
		c = _DeviceEntry(c)
		_configuration[device_key] = c

	return c


def attach_to(device):
	"""Apply the last saved configuration to a device."""
	if not _configuration:
		_load()

	persister = _device_entry(device)
	device.persister = persister



# -*- python-mode -*-
# -*- coding: UTF-8 -*-

## Copyright (C) 2012-2013  Daniel Pavel
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from __future__ import absolute_import, division, print_function, unicode_literals

import errno as _errno

from logging import getLogger, INFO as _INFO
_log = getLogger(__name__)
del getLogger


from .i18n import _
from . import base as _base
from . import hidpp10 as _hidpp10
from . import hidpp20 as _hidpp20
from .common import strhex as _strhex
from .descriptors import DEVICES as _DESCRIPTORS
from .settings_templates import check_feature_settings as _check_feature_settings
from .base_usb import product_information as _product_information

_R = _hidpp10.REGISTERS

#
#
#

class PairedDevice(object):
	def __init__(self, receiver, number, link_notification=None):
		assert receiver
		self.receiver = receiver

		assert number > 0 and number <= receiver.max_devices
		# Device number, 1..6 for unifying devices, 1 otherwise.
		self.number = number
		# 'device active' flag; requires manual management.
		self.online = None

		# the Wireless PID is unique per device model
		self.wpid = None
		self.descriptor = None

		# mouse, keyboard, etc (see _hidpp10.DEVICE_KIND)
		self._kind = None
		# Unifying peripherals report a codename.
		self._codename = None
		# the full name of the model
		self._name = None
		# HID++ protocol version, 1.0 or 2.0
		self._protocol = None
		# serial number (an 8-char hex string)
		self._serial = None

		self._firmware = None
		self._keys = None
		self._registers = None
		self._settings = None

		# Misc stuff that's irrelevant to any functionality, but may be
		# displayed in the UI and caching it here helps.
		self._polling_rate = None
		self._power_switch = None

		# if _log.isEnabledFor(_DEBUG):
		# 	_log.debug("new PairedDevice(%s, %s, %s)", receiver, number, link_notification)

		if link_notification is not None:
			self.online = not bool(ord(link_notification.data[0:1]) & 0x40)
			self.wpid = _strhex(link_notification.data[2:3] + link_notification.data[1:2])
			# assert link_notification.address == (0x04 if unifying else 0x03)
			kind = ord(link_notification.data[0:1]) & 0x0F
			self._kind = _hidpp10.DEVICE_KIND[kind]
		else:
			# force a reading of the wpid
			pair_info = receiver.read_register(_R.receiver_info, 0x20 + number - 1)
			if pair_info:
				# may be either a Unifying receiver, or an Unifying-ready receiver
				self.wpid = _strhex(pair_info[3:5])
				kind = ord(pair_info[7:8]) & 0x0F
				self._kind = _hidpp10.DEVICE_KIND[kind]
				self._polling_rate = ord(pair_info[2:3])

			else:
				# unifying protocol not supported, must be a Nano receiver
				device_info = self.receiver.read_register(_R.receiver_info, 0x04)
				if device_info is None:
					_log.error("failed to read Nano wpid for device %d of %s", number, receiver)
					raise _base.NoSuchDevice(number=number, receiver=receiver, error="read Nano wpid")

				self.wpid = _strhex(device_info[3:5])
				self._polling_rate = 0
				self._power_switch = '(' + _("unknown") + ')'

		# the wpid is necessary to properly identify wireless link on/off notifications
		# also it gets set to None on this object when the device is unpaired
		assert self.wpid is not None, "failed to read wpid: device %d of %s" % (number, receiver)

		self.descriptor = _DESCRIPTORS.get(self.wpid)
		if self.descriptor is None:
			# Last chance to correctly identify the device; many Nano receivers
			# do not support this call.
			codename = self.receiver.read_register(_R.receiver_info, 0x40 + self.number - 1)
			if codename:
				codename_length = ord(codename[1:2])
				codename = codename[2:2 + codename_length]
				self._codename = codename.decode('ascii')
				self.descriptor = _DESCRIPTORS.get(self._codename)

		if self.descriptor:
			self._name = self.descriptor.name
			self._protocol = self.descriptor.protocol
			if self._codename is None:
				self._codename = self.descriptor.codename
			if self._kind is None:
				self._kind = self.descriptor.kind

		if self._protocol is not None:
			self.features = None if self._protocol < 2.0 else _hidpp20.FeaturesArray(self)
		else:
			# may be a 2.0 device; if not, it will fix itself later
			self.features = _hidpp20.FeaturesArray(self)

	@property
	def protocol(self):
		if self._protocol is None and self.online is not False:
			self._protocol = _base.ping(self.receiver.handle, self.number)
			# if the ping failed, the peripheral is (almost) certainly offline
			self.online = self._protocol is not None

			# if _log.isEnabledFor(_DEBUG):
			# 	_log.debug("device %d protocol %s", self.number, self._protocol)
		return self._protocol or 0

	@property
	def codename(self):
		if self._codename is None:
			codename = self.receiver.read_register(_R.receiver_info, 0x40 + self.number - 1)
			if codename:
				codename_length = ord(codename[1:2])
				codename = codename[2:2 + codename_length]
				self._codename = codename.decode('ascii')
				# if _log.isEnabledFor(_DEBUG):
				#	 _log.debug("device %d codename %s", self.number, self._codename)
			else:
				self._codename = '? (%s)' % self.wpid
		return self._codename

	@property
	def name(self):
		if self._name is None:
			if self.online and self.protocol >= 2.0:
				self._name = _hidpp20.get_name(self)
		return self._name or self.codename  or ('Unknown device %s' % self.wpid)

	@property
	def kind(self):
		if self._kind is None:
			pair_info = self.receiver.read_register(_R.receiver_info, 0x20 + self.number - 1)
			if pair_info:
				kind = ord(pair_info[7:8]) & 0x0F
				self._kind = _hidpp10.DEVICE_KIND[kind]
				self._polling_rate = ord(pair_info[2:3])
			elif self.online and self.protocol >= 2.0:
				self._kind = _hidpp20.get_kind(self)
		return self._kind or '?'

	@property
	def firmware(self):
		if self._firmware is None and self.online:
			if self.protocol >= 2.0:
				self._firmware = _hidpp20.get_firmware(self)
			else:
				self._firmware = _hidpp10.get_firmware(self)
		return self._firmware or ()

	@property
	def serial(self):
		if self._serial is None:
			serial = self.receiver.read_register(_R.receiver_info, 0x30 + self.number - 1)
			if serial:
				ps = ord(serial[9:10]) & 0x0F
				self._power_switch = _hidpp10.POWER_SWITCH_LOCATION[ps]
			else:
				# some Nano receivers?
				serial = self.receiver.read_register(0x2D5)

			if serial:
				self._serial = _strhex(serial[1:5])
			else:
				# fallback...
				self._serial = self.receiver.serial
		return self._serial or '?'

	@property
	def power_switch_location(self):
		if self._power_switch is None:
			ps = self.receiver.read_register(_R.receiver_info, 0x30 + self.number - 1)
			if ps is not None:
				ps = ord(ps[9:10]) & 0x0F
				self._power_switch = _hidpp10.POWER_SWITCH_LOCATION[ps]
			else:
				self._power_switch = '(unknown)'
		return self._power_switch

	@property
	def polling_rate(self):
		if self._polling_rate is None:
			pair_info = self.receiver.read_register(_R.receiver_info, 0x20 + self.number - 1)
			if pair_info:
				self._polling_rate = ord(pair_info[2:3])
			else:
				self._polling_rate = 0
		return self._polling_rate

	@property
	def keys(self):
		if self._keys is None:
			if self.online and self.protocol >= 2.0:
				self._keys = _hidpp20.get_keys(self) or ()
		return self._keys

	@property
	def registers(self):
		if self._registers is None:
			if self.descriptor and self.descriptor.registers:
				self._registers = list(self.descriptor.registers)
			else:
				self._registers = []
		return self._registers

	@property
	def settings(self):
		if self._settings is None:
			if self.descriptor and self.descriptor.settings:
				self._settings = [s(self) for s in self.descriptor.settings]
			else:
				self._settings = []

		_check_feature_settings(self, self._settings)
		return self._settings

	def enable_notifications(self, enable=True):
		"""Enable or disable device (dis)connection notifications on this
		receiver."""
		if not bool(self.receiver) or self.protocol >= 2.0:
			return False

		if enable:
			set_flag_bits = ( _hidpp10.NOTIFICATION_FLAG.battery_status
							| _hidpp10.NOTIFICATION_FLAG.keyboard_illumination
							| _hidpp10.NOTIFICATION_FLAG.wireless
							| _hidpp10.NOTIFICATION_FLAG.software_present )
		else:
			set_flag_bits = 0
		ok = _hidpp10.set_notification_flags(self, set_flag_bits)
		if ok is None:
			_log.warn("%s: failed to %s device notifications", self, 'enable' if enable else 'disable')

		flag_bits = _hidpp10.get_notification_flags(self)
		flag_names = None if flag_bits is None else tuple(_hidpp10.NOTIFICATION_FLAG.flag_names(flag_bits))
		if _log.isEnabledFor(_INFO):
			_log.info("%s: device notifications %s %s", self, 'enabled' if enable else 'disabled', flag_names)
		return flag_bits if ok else None

	def request(self, request_id, *params):
		return _base.request(self.receiver.handle, self.number, request_id, *params)

	read_register = _hidpp10.read_register
	write_register = _hidpp10.write_register

	def feature_request(self, feature, function=0x00, *params):
		if self.protocol >= 2.0:
			return _hidpp20.feature_request(self, feature, function, *params)

	def ping(self):
		"""Checks if the device is online, returns True of False"""
		protocol = _base.ping(self.receiver.handle, self.number)
		self.online = protocol is not None
		if protocol is not None:
			self._protocol = protocol
		return self.online

	def __index__(self):
		return self.number
	__int__ = __index__

	def __eq__(self, other):
		return other is not None and self.kind == other.kind and self.wpid == other.wpid

	def __ne__(self, other):
		return other is None or self.kind != other.kind or self.wpid != other.wpid

	def __hash__(self):
		return self.wpid.__hash__()

	__bool__ = __nonzero__ = lambda self: self.wpid is not None and self.number in self.receiver

	def __str__(self):
		return '<PairedDevice(%d,%s,%s,%s)>' % (self.number, self.wpid, self.codename or '?', self.serial)
	__unicode__ = __repr__ = __str__

#
#
#

class Receiver(object):
	"""A Unifying Receiver instance.

	The paired devices are available through the sequence interface.
	"""
	number = 0xFF
	kind = None

	def __init__(self, handle, device_info):
		assert handle
		self.handle = handle
		assert device_info
		self.path = device_info.path
		# USB product id, used for some Nano receivers
		self.product_id = device_info.product_id
		product_info = _product_information(self.product_id)
		if not product_info:
			raise Exception("Unknown receiver type", self.product_id)

		# read the serial immediately, so we can find out max_devices
		serial_reply = self.read_register(_R.receiver_info, 0x03)
		if serial_reply :
			self.serial = _strhex(serial_reply[1:5])
			self.max_devices = ord(serial_reply[6:7])
			# TODO _properly_ figure out which receivers do and which don't support unpairing
			# This code supposes that receivers that don't unpair support a pairing request for device index 0
			self.may_unpair = self.write_register(_R.receiver_pairing) is None
		else: # handle receivers that don't have a serial number specially (i.e., c534)
			self.serial = None
			self.max_devices = product_info.get('max_devices',1)
			self.may_unpair = product_info.get('may_unpair',False)

		self.name = product_info.get('name','')
		self.re_pairs = product_info.get('re_pairs',False)
		self._str = '<%s(%s,%s%s)>' % (self.name.replace(' ', ''), self.path, '' if isinstance(self.handle, int) else 'T', self.handle)

		self._firmware = None
		self._devices = {}
		self._remaining_pairings = None

	def close(self):
		handle, self.handle = self.handle, None
		self._devices.clear()
		return (handle and _base.close(handle))

	def __del__(self):
		self.close()

	@property
	def firmware(self):
		if self._firmware is None and self.handle:
			self._firmware = _hidpp10.get_firmware(self)
		return self._firmware

	# how many pairings remain (None for unknown, -1 for unlimited)
	def remaining_pairings(self,cache=True):
		if self._remaining_pairings is None or not cache:
			ps = self.read_register(_R.receiver_connection)
			if ps is not None:
				ps = ord(ps[2:3])
				self._remaining_pairings = ps-5 if ps >= 5 else -1
		return self._remaining_pairings

	def enable_notifications(self, enable=True):
		"""Enable or disable device (dis)connection notifications on this
		receiver."""
		if not self.handle:
			return False

		if enable:
			set_flag_bits = ( _hidpp10.NOTIFICATION_FLAG.battery_status
							| _hidpp10.NOTIFICATION_FLAG.wireless
							| _hidpp10.NOTIFICATION_FLAG.software_present )
		else:
			set_flag_bits = 0
		ok = _hidpp10.set_notification_flags(self, set_flag_bits)
		if ok is None:
			_log.warn("%s: failed to %s receiver notifications", self, 'enable' if enable else 'disable')
			return None

		flag_bits = _hidpp10.get_notification_flags(self)
		flag_names = None if flag_bits is None else tuple(_hidpp10.NOTIFICATION_FLAG.flag_names(flag_bits))
		if _log.isEnabledFor(_INFO):
			_log.info("%s: receiver notifications %s => %s", self, 'enabled' if enable else 'disabled', flag_names)
		return flag_bits

	def notify_devices(self):
		"""Scan all devices."""
		if self.handle:
			if not self.write_register(_R.receiver_connection, 0x02):
				_log.warn("%s: failed to trigger device link notifications", self)

	def register_new_device(self, number, notification=None):
		if self._devices.get(number) is not None:
			raise IndexError("%s: device number %d already registered" % (self, number))

		assert notification is None or notification.devnumber == number
		assert notification is None or notification.sub_id == 0x41

		try:
			dev = PairedDevice(self, number, notification)
			assert dev.wpid
			if _log.isEnabledFor(_INFO):
				_log.info("%s: found new device %d (%s)", self, number, dev.wpid)
			self._devices[number] = dev
			return dev
		except _base.NoSuchDevice:
			_log.exception("register_new_device")

		_log.warning("%s: looked for device %d, not found", self, number)
		self._devices[number] = None

	def set_lock(self, lock_closed=True, device=0, timeout=0):
		if self.handle:
			action = 0x02 if lock_closed else 0x01
			reply = self.write_register(_R.receiver_pairing, action, device, timeout)
			if reply:
				return True
			_log.warn("%s: failed to %s the receiver lock", self, 'close' if lock_closed else 'open')

	def count(self):
		count = self.read_register(_R.receiver_connection)
		return 0 if count is None else ord(count[1:2])

	# def has_devices(self):
	# 	return len(self) > 0 or self.count() > 0

	def request(self, request_id, *params):
		if bool(self):
			return _base.request(self.handle, 0xFF, request_id, *params)

	read_register = _hidpp10.read_register
	write_register = _hidpp10.write_register

	def __iter__(self):
		for number in range(1, 1 + self.max_devices):
			if number in self._devices:
				dev = self._devices[number]
			else:
				dev = self.__getitem__(number)
			if dev is not None:
				yield dev

	def __getitem__(self, key):
		if not bool(self):
			return None

		dev = self._devices.get(key)
		if dev is not None:
			return dev

		if not isinstance(key, int):
			raise TypeError('key must be an integer')
		if key < 1 or key > self.max_devices:
			raise IndexError(key)

		return self.register_new_device(key)

	def __delitem__(self, key):
		self._unpair_device(key, False)

	def _unpair_device(self, key, force=False):
		key = int(key)

		if self._devices.get(key) is None:
			raise IndexError(key)

		dev = self._devices[key]
		if not dev:
			if key in self._devices:
				del self._devices[key]
			return

		if self.re_pairs and not force:
			# invalidate the device, but these receivers don't unpair per se
			dev.online = False
			dev.wpid = None
			if key in self._devices:
				del self._devices[key]
			_log.warn("%s removed device %s", self, dev)
		else:
			reply = self.write_register(_R.receiver_pairing, 0x03, key)
			if reply:
				# invalidate the device
				dev.online = False
				dev.wpid = None
				if key in self._devices:
					del self._devices[key]
				_log.warn("%s unpaired device %s", self, dev)
			else:
				_log.error("%s failed to unpair device %s", self, dev)
				raise IndexError(key)

	def __len__(self):
		return len([d for d in self._devices.values() if d is not None])

	def __contains__(self, dev):
		if isinstance(dev, int):
			return self._devices.get(dev) is not None

		return self.__contains__(dev.number)

	def __eq__(self, other):
		return other is not None and self.kind == other.kind and self.path == other.path

	def __ne__(self, other):
		return other is None or self.kind != other.kind or self.path != other.path

	def __hash__(self):
		return self.path.__hash__()

	def __str__(self):
		return self._str
	__unicode__ = __repr__ = __str__

	__bool__ = __nonzero__ = lambda self: self.handle is not None

	@classmethod
	def open(self, device_info):
		"""Opens a Logitech Receiver found attached to the machine, by Linux device path.

		:returns: An open file handle for the found receiver, or ``None``.
		"""
		try:
			handle = _base.open_path(device_info.path)
			if handle:
				return Receiver(handle, device_info)
		except OSError as e:
			_log.exception("open %s", device_info)
			if e.errno == _errno.EACCES:
				raise
		except:
			_log.exception("open %s", device_info)



# -*- python-mode -*-
# -*- coding: UTF-8 -*-

## Copyright (C) 2012-2013  Daniel Pavel
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from __future__ import absolute_import, division, print_function, unicode_literals

__version__ = '1.0.2'
NAME = 'Solaar'



# -*- python-mode -*-
# -*- coding: UTF-8 -*-

## Copyright (C) 2012-2013  Daniel Pavel
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from __future__ import absolute_import, division, print_function, unicode_literals
import os

from logging import getLogger, DEBUG as _DEBUG
_log = getLogger(__name__)
del getLogger

from time import time as _timestamp

from gi.repository import Gtk, GLib
from gi.repository.Gdk import ScrollDirection


from solaar import NAME
from solaar.i18n import _
from logitech_receiver.status import KEYS as _K
from . import icons as _icons
from .window import popup as _window_popup, toggle as _window_toggle

#
# constants
#

_TRAY_ICON_SIZE = 32 #  pixels
_MENU_ICON_SIZE = Gtk.IconSize.LARGE_TOOLBAR
_RECEIVER_SEPARATOR = ('~', None, None, None)

#
#
#

def _create_menu(quit_handler):
	menu = Gtk.Menu()

	# per-device menu entries will be generated as-needed

	no_receiver = Gtk.MenuItem.new_with_label(_("No Logitech receiver found"))
	no_receiver.set_sensitive(False)
	menu.append(no_receiver)
	menu.append(Gtk.SeparatorMenuItem.new())

	from .action import about, make
	menu.append(about.create_menu_item())
	menu.append(make('application-exit', _("Quit"), quit_handler, stock_id=Gtk.STOCK_QUIT).create_menu_item())
	del about, make

	menu.show_all()

	return menu


_last_scroll = 0
def _scroll(tray_icon, event, direction=None):
	if direction is None:
		direction = event.direction
		now = event.time / 1000.0
	else:
		now = None

	if direction != ScrollDirection.UP and direction != ScrollDirection.DOWN:
		# ignore all other directions
		return

	if len(_devices_info) < 4:
		# don't bother with scrolling when there's only one receiver
		# with only one device (3 = [receiver, device, separator])
		return

	# scroll events come way too fast (at least 5-6 at once)
	# so take a little break between them
	global _last_scroll
	now = now or _timestamp()
	if now - _last_scroll < 0.33:  # seconds
		return
	_last_scroll = now

	# if _log.isEnabledFor(_DEBUG):
	# 	_log.debug("scroll direction %s", direction)

	global _picked_device
	candidate = None

	if _picked_device is None:
		for info in _devices_info:
			# pick first peripheral found
			if info[1] is not None:
				candidate = info
				break
	else:
		found = False
		for info in _devices_info:
			if not info[1]:
				# only conside peripherals
				continue
			# compare peripherals
			if info[0:2] == _picked_device[0:2]:
				if direction == ScrollDirection.UP and candidate:
					# select previous device
					break
				found = True
			else:
				if found:
					candidate = info
					if direction == ScrollDirection.DOWN:
						break
					# if direction is up, but no candidate found before _picked,
					# let it run through all candidates, will get stuck with the last one
				else:
					if direction == ScrollDirection.DOWN:
						# only use the first one, in case no candidates are after _picked
						if candidate is None:
							candidate = info
					else:
						candidate = info

		# if the last _picked_device is gone, clear it
		# the candidate will be either the first or last one remaining,
		# depending on the scroll direction
		if not found:
			_picked_device = None

	_picked_device = candidate or _picked_device
	if _log.isEnabledFor(_DEBUG):
		_log.debug("scroll: picked %s", _picked_device)
	_update_tray_icon()


try:
	import gi
	try:
		gi.require_version('AyatanaAppIndicator3', '0.1')
		ayatana_appindicator_found = True
	except ValueError:
		try:
			gi.require_version('AppIndicator3', '0.1')
			ayatana_appindicator_found = False
		except ValueError:
			# treat unavailable versions the same as unavailable packages
			raise ImportError

	if ayatana_appindicator_found:
		from gi.repository import AyatanaAppIndicator3 as AppIndicator3
	else:
		from gi.repository import AppIndicator3

	if _log.isEnabledFor(_DEBUG):
		_log.debug("using %sAppIndicator3" % ('Ayatana ' if ayatana_appindicator_found else ''))

	# Defense against AppIndicator3 bug that treats files in current directory as icon files
	# https://bugs.launchpad.net/ubuntu/+source/libappindicator/+bug/1363277
	def _icon_file(icon_name):
		if not os.path.isfile(icon_name):
			return icon_name
		icon_info = Gtk.IconTheme.get_default().lookup_icon(icon_name,_TRAY_ICON_SIZE,0)
		return icon_info.get_filename() if icon_info else icon_name

	def _create(menu):
		theme_paths = Gtk.IconTheme.get_default().get_search_path()

		ind = AppIndicator3.Indicator.new_with_path(
						'indicator-solaar',
						_icon_file(_icons.TRAY_INIT),
						AppIndicator3.IndicatorCategory.HARDWARE,
						':'.join(theme_paths))
		ind.set_title(NAME)
		ind.set_status(AppIndicator3.IndicatorStatus.ACTIVE)
		ind.set_attention_icon_full(_icon_file(_icons.TRAY_ATTENTION), '')
		# ind.set_label(NAME, NAME)

		ind.set_menu(menu)
		ind.connect('scroll-event', _scroll)

		return ind


	def _destroy(indicator):
		indicator.set_status(AppIndicator3.IndicatorStatus.PASSIVE)


	def _update_tray_icon():
		if _picked_device:
			_ignore, _ignore, name, device_status = _picked_device
			battery_level = device_status.get(_K.BATTERY_LEVEL)
			battery_charging = device_status.get(_K.BATTERY_CHARGING)
			tray_icon_name = _icons.battery(battery_level, battery_charging)

			description =  '%s: %s' % (name, device_status.to_string())
		else:
			# there may be a receiver, but no peripherals
			tray_icon_name = _icons.TRAY_OKAY if _devices_info else _icons.TRAY_INIT

			description_lines = _generate_description_lines()
			description = '\n'.join(description_lines).rstrip('\n')

		# icon_file = _icons.icon_file(icon_name, _TRAY_ICON_SIZE)
		_icon.set_icon_full(_icon_file(tray_icon_name), description)


	def _update_menu_icon(image_widget, icon_name):
		image_widget.set_from_icon_name(icon_name, _MENU_ICON_SIZE)
		# icon_file = _icons.icon_file(icon_name, _MENU_ICON_SIZE)
		# image_widget.set_from_file(icon_file)
		# image_widget.set_pixel_size(_TRAY_ICON_SIZE)


	def attention(reason=None):
		if _icon.get_status() != AppIndicator3.IndicatorStatus.ATTENTION:
			_icon.set_attention_icon_full(_icon_file(_icons.TRAY_ATTENTION), reason or '')
			_icon.set_status(AppIndicator3.IndicatorStatus.ATTENTION)
			GLib.timeout_add(10 * 1000, _icon.set_status, AppIndicator3.IndicatorStatus.ACTIVE)

except ImportError:

	if _log.isEnabledFor(_DEBUG):
		_log.debug("using StatusIcon")

	def _create(menu):
		icon = Gtk.StatusIcon.new_from_icon_name(_icons.TRAY_INIT)
		icon.set_name(NAME)
		icon.set_title(NAME)
		icon.set_tooltip_text(NAME)
		icon.connect('activate', _window_toggle)
		icon.connect('scroll-event', _scroll)
		icon.connect('popup-menu',
						lambda icon, button, time: menu.popup(None, None, icon.position_menu, icon, button, time))

		return icon


	def _destroy(icon):
		icon.set_visible(False)


	def _update_tray_icon():
		tooltip_lines = _generate_tooltip_lines()
		tooltip = '\n'.join(tooltip_lines).rstrip('\n')
		_icon.set_tooltip_markup(tooltip)

		if _picked_device:
			_ignore, _ignore, name, device_status = _picked_device
			battery_level = device_status.get(_K.BATTERY_LEVEL)
			battery_charging = device_status.get(_K.BATTERY_CHARGING)
			tray_icon_name = _icons.battery(battery_level, battery_charging)
		else:
			# there may be a receiver, but no peripherals
			tray_icon_name = _icons.TRAY_OKAY if _devices_info else _icons.TRAY_ATTENTION
		_icon.set_from_icon_name(tray_icon_name)


	def _update_menu_icon(image_widget, icon_name):
		image_widget.set_from_icon_name(icon_name, _MENU_ICON_SIZE)


	_icon_before_attention = None

	def _blink(count):
		global _icon_before_attention
		if count % 2:
			_icon.set_from_icon_name(_icons.TRAY_ATTENTION)
		else:
			_icon.set_from_icon_name(_icon_before_attention)

		if count > 0:
			GLib.timeout_add(1000, _blink, count - 1)
		else:
			_icon_before_attention = None

	def attention(reason=None):
		global _icon_before_attention
		if _icon_before_attention is None:
			_icon_before_attention = _icon.get_icon_name()
			GLib.idle_add(_blink, 9)

#
#
#

def _generate_tooltip_lines():
	if not _devices_info:
		yield '<b>%s</b>: ' % NAME + _("no receiver")
		return

	yield from _generate_description_lines()


def _generate_description_lines():
	if not _devices_info:
		yield _("no receiver")
		return

	for _ignore, number, name, status in _devices_info:
		if number is None:  # receiver
			continue

		p = status.to_string()
		if p:  # does it have any properties to print?
			yield '<b>%s</b>' % name
			if status:
				yield '\t%s' % p
			else:
				yield '\t%s <small>(' % p + _("offline") + ')</small>'
		else:
			if status:
				yield '<b>%s</b> <small>(' % name + _("no status") + ')</small>'
			else:
				yield '<b>%s</b> <small>(' % name + _("offline") + ')</small>'
		yield ''


def _pick_device_with_lowest_battery():
	if not _devices_info:
		return None

	picked = None
	picked_level = 1000

	for info in _devices_info:
		if info[1] is None: # is receiver/separator
			continue
		level = info[-1].get(_K.BATTERY_LEVEL)
		# print ("checking %s -> %s", info, level)
		if level is not None and picked_level > level:
			picked = info
			picked_level = level or 0

	if _log.isEnabledFor(_DEBUG):
		_log.debug("picked device with lowest battery: %s", picked)

	return picked


#
#
#

def _add_device(device):
	assert device
	assert device.receiver
	receiver_path = device.receiver.path
	assert receiver_path

	index = None
	for idx, (path, _ignore, _ignore, _ignore) in enumerate(_devices_info):
		if path == receiver_path:
			# the first entry matching the receiver serial should be for the receiver itself
			index = idx + 1
			break
	assert index is not None

	# proper ordering (according to device.number) for a receiver's devices
	while True:
		path, number, _ignore, _ignore = _devices_info[index]
		if path == _RECEIVER_SEPARATOR[0]:
			break
		assert path == receiver_path
		assert number != device.number
		if number > device.number:
			break
		index = index + 1

	new_device_info = (receiver_path, device.number, device.name, device.status)
	assert len(new_device_info) == len(_RECEIVER_SEPARATOR)
	_devices_info.insert(index, new_device_info)

	# label_prefix = b'\xE2\x94\x84 '.decode('utf-8')
	label_prefix = '   '

	new_menu_item = Gtk.ImageMenuItem.new_with_label(label_prefix + device.name)
	new_menu_item.set_image(Gtk.Image())
	new_menu_item.show_all()
	new_menu_item.connect('activate', _window_popup, receiver_path, device.number)
	_menu.insert(new_menu_item, index)

	return index


def _remove_device(index):
	assert index is not None

	menu_items = _menu.get_children()
	_menu.remove(menu_items[index])

	removed_device = _devices_info.pop(index)
	global _picked_device
	if _picked_device and _picked_device[0:2] == removed_device[0:2]:
		# the current pick was unpaired
		_picked_device = None


def _add_receiver(receiver):
	index = len(_devices_info)

	new_receiver_info = (receiver.path, None, receiver.name, None)
	assert len(new_receiver_info) == len(_RECEIVER_SEPARATOR)
	_devices_info.append(new_receiver_info)

	new_menu_item = Gtk.ImageMenuItem.new_with_label(receiver.name)
	_menu.insert(new_menu_item, index)
	icon_set = _icons.device_icon_set(receiver.name)
	new_menu_item.set_image(Gtk.Image().new_from_icon_set(icon_set, _MENU_ICON_SIZE))
	new_menu_item.show_all()
	new_menu_item.connect('activate', _window_popup, receiver.path)

	_devices_info.append(_RECEIVER_SEPARATOR)
	separator = Gtk.SeparatorMenuItem.new()
	separator.set_visible(True)
	_menu.insert(separator, index + 1)

	return 0


def _remove_receiver(receiver):
	index = 0
	found = False

	# remove all entries in devices_info that match this receiver
	while index < len(_devices_info):
		path, _ignore, _ignore, _ignore = _devices_info[index]
		if path == receiver.path:
			found = True
			_remove_device(index)
		elif found and path == _RECEIVER_SEPARATOR[0]:
			# the separator after this receiver
			_remove_device(index)
			break
		else:
			index += 1


def _update_menu_item(index, device):
	assert device
	assert device.status is not None

	menu_items = _menu.get_children()
	menu_item = menu_items[index]

	level = device.status.get(_K.BATTERY_LEVEL)
	charging = device.status.get(_K.BATTERY_CHARGING)
	icon_name = _icons.battery(level, charging)

	image_widget = menu_item.get_image()
	image_widget.set_sensitive(bool(device.online))
	_update_menu_icon(image_widget, icon_name)

#
#
#

# for which device to show the battery info in systray, if more than one
# it's actually an entry in _devices_info
_picked_device = None

# cached list of devices and some of their properties
# contains tuples of (receiver path, device number, name, status)
_devices_info = []

_menu = None
_icon = None

def init(_quit_handler):
	global _menu, _icon
	assert _menu is None
	_menu = _create_menu(_quit_handler)
	assert _icon is None
	_icon = _create(_menu)


def destroy():
	global _icon, _menu, _devices_info
	if _icon is not None:
		i, _icon = _icon, None
		_destroy(i)
		i = None

	_icon = None
	_menu = None
	_devices_info = None


def update(device=None):
	if _icon is None:
		return

	if device is not None:
		if device.kind is None:
			# receiver
			is_alive = bool(device)
			receiver_path = device.path
			if is_alive:
				index = None
				for idx, (path, _ignore, _ignore, _ignore) in enumerate(_devices_info):
					if path == receiver_path:
						index = idx
						break

				if index is None:
					_add_receiver(device)
			else:
				_remove_receiver(device)

		else:
			# peripheral
			is_paired = bool(device)
			receiver_path = device.receiver.path
			index = None
			for idx, (path, number, _ignore, _ignore) in enumerate(_devices_info):
				if path == receiver_path and number == device.number:
					index = idx

			if is_paired:
				if index is None:
					index = _add_device(device)
				_update_menu_item(index, device)
			else:
				# was just unpaired
				if index:
					_remove_device(index)

		menu_items = _menu.get_children()
		no_receivers_index = len(_devices_info)
		menu_items[no_receivers_index].set_visible(not _devices_info)
		menu_items[no_receivers_index + 1].set_visible(not _devices_info)

	global _picked_device
	if (not _picked_device or _last_scroll == 0) and device is not None and device.kind is not None:
		# if it's just a receiver update, it's unlikely the picked device would change
		_picked_device = _pick_device_with_lowest_battery()

	_update_tray_icon()



# -*- python-mode -*-
# -*- coding: UTF-8 -*-

## Copyright (C) 2012-2013  Daniel Pavel
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from __future__ import absolute_import, division, print_function, unicode_literals


from logging import getLogger, DEBUG as _DEBUG
_log = getLogger(__name__)
del getLogger

from gi.repository import GLib, Gtk


from solaar.i18n import _

#
#
#

assert Gtk.get_major_version() > 2, 'Solaar requires Gtk 3 python bindings'

GLib.threads_init()

#
#
#

def _error_dialog(reason, object):
	_log.error("error: %s %s", reason, object)

	if reason == 'permissions':
		title = _("Permissions error")
		text = _("Found a Logitech Receiver (%s), but did not have permission to open it.") % object + \
				'\n\n' + \
				_("If you've just installed Solaar, try removing the receiver and plugging it back in.")
	elif reason == 'unpair':
		title = _("Unpairing failed")
		text = _("Failed to unpair %{device} from %{receiver}.").format(device=object.name, receiver=object.receiver.name) + \
				'\n\n' + \
				_("The receiver returned an error, with no further details.")
	else:
		raise Exception("ui.error_dialog: don't know how to handle (%s, %s)", reason, object)

	assert title
	assert text

	m = Gtk.MessageDialog(None, Gtk.DialogFlags.MODAL, Gtk.MessageType.ERROR, Gtk.ButtonsType.CLOSE, text)
	m.set_title(title)
	m.run()
	m.destroy()


def error_dialog(reason, object):
	assert reason is not None
	GLib.idle_add(_error_dialog, reason, object)

#
#
#

_task_runner = None
def ui_async(function, *args, **kwargs):
	if _task_runner:
		_task_runner(function, *args, **kwargs)

#
#
#

from . import notify, tray, window


def _startup(app, startup_hook, use_tray, show_window):
	if _log.isEnabledFor(_DEBUG):
		_log.debug("startup registered=%s, remote=%s", app.get_is_registered(), app.get_is_remote())

	from solaar.tasks import TaskRunner as _TaskRunner
	global _task_runner
	_task_runner = _TaskRunner('AsyncUI')
	_task_runner.start()

	notify.init()
	if use_tray:
		tray.init(lambda _ignore: window.destroy())
	window.init(show_window, use_tray)

	startup_hook()


def _activate(app):
	if _log.isEnabledFor(_DEBUG):
		_log.debug("activate")
	if app.get_windows():
		window.popup()
	else:
		app.add_window(window._window)


def _command_line(app, command_line):
	if _log.isEnabledFor(_DEBUG):
		_log.debug("command_line %s", command_line.get_arguments())

	return 0


def _shutdown(app, shutdown_hook):
	if _log.isEnabledFor(_DEBUG):
		_log.debug("shutdown")

	shutdown_hook()

	# stop the async UI processor
	global _task_runner
	_task_runner.stop()
	_task_runner = None

	tray.destroy()
	notify.uninit()


def run_loop(startup_hook, shutdown_hook, use_tray, show_window, args=None):
	assert use_tray or show_window, 'need either tray or visible window'
	# from gi.repository.Gio import ApplicationFlags as _ApplicationFlags
	APP_ID = 'io.github.pwr.solaar'
	application = Gtk.Application.new(APP_ID, 0) # _ApplicationFlags.HANDLES_COMMAND_LINE)

	application.connect('startup', lambda app, startup_hook:_startup(app,startup_hook,use_tray,show_window), startup_hook)
	application.connect('command-line', _command_line)
	application.connect('activate', _activate)
	application.connect('shutdown', _shutdown, shutdown_hook)

	application.run(args)

#
#
#

from logitech_receiver.status import ALERT


def _status_changed(device, alert, reason):
	assert device is not None
	if _log.isEnabledFor(_DEBUG):
		_log.debug("status changed: %s (%s) %s", device, alert, reason)

	tray.update(device)
	if alert & ALERT.ATTENTION:
		tray.attention(reason)

	need_popup = alert & ALERT.SHOW_WINDOW
	window.update(device, need_popup)

	if alert & (ALERT.NOTIFICATION | ALERT.ATTENTION):
		notify.show(device, reason)


def status_changed(device, alert=ALERT.NONE, reason=None):
	GLib.idle_add(_status_changed, device, alert, reason)



# -*- python-mode -*-
# -*- coding: UTF-8 -*-

## Copyright (C) 2012-2013  Daniel Pavel
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.
##
## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from __future__ import absolute_import, division, print_function, unicode_literals


def run(receivers, args, find_receiver, find_device):
	assert receivers
	assert args.device

	device_name = args.device.lower()
	dev = find_device(receivers, device_name)

	if not dev.receiver.may_unpair:
		print('Receiver for %s [%s:%s] does not unpair, but attempting anyway' % (dev.name,dev.wpid,dev.serial))

	try:
		# query these now, it's last chance to get them
		number, codename, wpid, serial  = dev.number, dev.codename, dev.wpid, dev.serial
		dev.receiver._unpair_device(number, True) # force an unpair
		print ('Unpaired %d: %s (%s) [%s:%s]' % (number, dev.name, codename, wpid, serial))
	except Exception as e:
		raise Exception('failed to unpair device %s: %s' % (dev.name, e))
